/*
  FORTRAN callable routines
*/

#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>

#if defined WINTEL
#include <windows.h>
#else
#include <unistd.h>
#endif

#include <string.h>

#if !defined WINTEL
int sched_getcpu();
#endif

#if defined WINTEL
void SLEEP_F(long int *time)
{
   DWORD time_ms = (*time) * 1000;
   Sleep(time_ms);
// On windows Sleep() does not return anything so there is no way to check if it has been interrupted
}
#else


#ifdef INT64
void sleep_f_(long int *time)
#else
void sleep_f_(int *time)
#endif
{
   unsigned int time_rem;
   time_rem = sleep((unsigned int)*time);
   if(time_rem != 0) {
      printf("Sleep interrupted by signal handler.\n");
   }
}

#endif

#if defined WINTEL
void USLEEP_F(long int *time)
{
   DWORD time_ms = ((*time) - ((*time) % 1000))/1000 + 1;
   Sleep(time_ms);
}
#else

#ifdef INT64
void usleep_f_(long int *time)
#else
void usleep_f_(int *time)
#endif
{
   int error;
   error = usleep((useconds_t)*time);
   if(error != 0) {
      if(errno == EINTR) {
         printf("Sleep interrupted by a signal.\n");
      } else if(errno == EINVAL) {
         usleep((useconds_t)1000000);
      }
   }
}
#endif

#if defined WINTEL
long int GETCPU ()
#elif defined INT64
long int getcpu_ ()
#else
int getcpu_ ()
#endif
{
   int cpu;
#if defined WINTEL
   cpu = GetCurrentProcessorNumber();
#else
   cpu = sched_getcpu();
#endif
   return cpu;
}

#if defined WINTEL
long int GETPID_F () {
   long int pid;
#elif defined INT64
long int getpid_f_ () {
   long int pid;
#else
int getpid_f_ () {
   int pid;
#endif
   pid = getpid();
   return pid;
}

#if defined WINTEL
void EXEC_MPIRUN(int *implementation) {
#else
void exec_mpirun_(int *implementation) {
#endif
   int nargs = 5 + (*implementation == 2 ? 3 : 0);
   char **args = (char **)calloc(nargs, sizeof(char *));
#if defined WINTEL
   args[0] = "mpiexec";
#else
   args[0] = "mpirun";
#endif
   args[1] = "-np";
   args[2] = "1";
   if(*implementation == 2) {
      args[3] = "-oversubscribe";
      args[4] = "-map-by";
      args[5] = "node";
   }
   args[nargs - 2] = "dmrcc_mpi";
#if defined WINTEL
   execvp("mpiexec", args);
#else
   execvp("mpirun", args);
#endif
}

#ifdef STATIC
asm (".symver memcpy, memcpy@GLIBC_2.2.5");

void *__wrap_memcpy(void *dest, const void *src, size_t n)
{
   return memcpy(dest, src, n);
}
#endif

#ifdef __cplusplus
}
#endif
