#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#if defined WINTEL
#include <windows.h>
#include <process.h>
typedef int pid_t;
#else
#include <unistd.h>
#endif


#ifdef __cplusplus
extern "C" {
#endif

#if defined WINTEL
void MRCCEND(long*);
void DMRCCEND(long*);
long ERROR_SIGTERM = -1;
long ERROR_SIGSEGV = -2;
#elif defined INT64
void mrccend_(long*);
void dmrccend_(long*);
long ERROR_SIGTERM = -1;
long ERROR_SIGSEGV = -2;
#else
void mrccend_(int*);
void dmrccend_(int*);
int ERROR_SIGTERM = -1;
int ERROR_SIGSEGV = -2;
#endif

#if defined WINTEL
static void (*old_segv_handler)(int) = NULL;
void PARENT_SIGTERM(int);
void CHILD_SIGTERM(int);
void CHILD_SIGSEGV(int);
#else
void parent_sigterm_(int, siginfo_t *, void *);
void child_sigterm_(int, siginfo_t *, void *);
void child_sigsegv_(int, siginfo_t *, void *);
#endif
void sendSignalToChildren(int);

#if defined WINTEL
#else
static struct sigaction old_act;
#endif

#if defined WINTEL
void SIGNALINIT() {
   signal(SIGTERM, PARENT_SIGTERM);
}
#else
void signalinit_() {
// initialise dmrcc's responses to signals
   struct sigaction act_term;
   memset(&act_term, '\0', sizeof(act_term));
   act_term.sa_sigaction = &parent_sigterm_;
   act_term.sa_flags = SA_SIGINFO;
   sigaction(SIGTERM, &act_term, NULL);
}
#endif

#if defined WINTEL
void PARENT_SIGTERM(int signum) {
#else
void parent_sigterm_(int signum, siginfo_t *siginfo, void *context) {
#endif
// initialise response to signal SIGTERM
   pid_t pid_parent;
   char pidchar[10];
   char command[40];

   pid_parent = getpid();

   sprintf(pidchar, "%d", pid_parent);

   printf("\n Program dmrcc recieved SIGTERM\n"); fflush(stdout);

   sendSignalToChildren(SIGTERM);
   Sleep(5);
#if !defined WINTEL
   sendSignalToChildren(SIGKILL);
#endif

   printf("\n Program dmrcc terminating\n"); fflush(stdout);


#if defined WINTEL
   DMRCCEND(&ERROR_SIGTERM);
#else
   dmrccend_(&ERROR_SIGTERM);
#endif
}

void sendSignalToChildren(int sig) {
   int ownPid = _getpid();
   int pid, numPids = 0;
   int *pids;
#if defined WINTEL
   char killfn[40];
   char pidstring[15];
#endif
   FILE *pidfile = fopen("pids", "r");
   if (pidfile == NULL) {
      printf("Error: Could not open pids file\n");
      return;
   }

   // number of running processes other than the current process
   while (fscanf(pidfile, "%d", &pid) != EOF) {
      if (pid != ownPid) {
         numPids++;
      }
   }
   rewind(pidfile);

   // read other process' PIDs
   pids = (int *)malloc(numPids * sizeof(int));
   int i = -1;
   while (fscanf(pidfile, "%d", &pid) != EOF) {
      if (pid != ownPid) {
         pids[++i] = pid;
      }
   }


   // send signal sig to processes
   printf("\n Sending signal %2d to child processes\n", sig); fflush(stdout);
   for (i = 0; i < numPids; i++) {

#if defined WINTEL
      sprintf(pidstring, "%d", pids[i]);
      sprintf(killfn,"%s","taskkill /F /T /PID ");
      strcat(killfn,pidstring);
      system(killfn);
#else
      kill((pid_t)pids[i], sig);
#endif
   }
   
   fclose(pidfile);
   free(pids);
}
#if defined WINTEL
void SIGNALINITCHILD() {
   signal(SIGTERM, CHILD_SIGTERM);
   old_segv_handler = signal(SIGSEGV, CHILD_SIGSEGV);
}
#else
void signalinitchild_() {
// initialise child's responses to signals
   struct sigaction act_term;
   memset (&act_term, '\0', sizeof(act_term));
   act_term.sa_sigaction = &child_sigterm_;
   act_term.sa_flags = SA_SIGINFO;
   sigaction(SIGTERM, &act_term, NULL);

   struct sigaction act_segv;
   memset (&act_segv, '\0', sizeof(act_segv));
   act_segv.sa_sigaction = &child_sigsegv_;
   act_segv.sa_flags = SA_SIGINFO;
   sigaction(SIGSEGV, &act_segv, &old_act);
}
#endif

#if defined WINTEL
void CHILD_SIGTERM(int signum) {
   MRCCEND(&ERROR_SIGTERM);
}
#else
void child_sigterm_(int signum, siginfo_t *siginfo, void *context) {
// initialise child's response to signal SIGTERM
   mrccend_(&ERROR_SIGTERM);
}
#endif

#if defined WINTEL
void CHILD_SIGSEGV(int signum) {
   MRCCEND(&ERROR_SIGSEGV);
   signal(SIGSEGV,old_segv_handler);
   raise(SIGSEGV);
}
#else
void child_sigsegv_(int signum, siginfo_t *siginfo, void *context) {
// initialise child's response to signal SIGSEGV
   mrccend_(&ERROR_SIGSEGV);
   if(old_act.sa_flags & SA_SIGINFO)
   {
      (*old_act.sa_sigaction)(signum, siginfo, context);
   } 
   else 
   { 
      (*old_act.sa_handler)(signum);
   }
}
#endif
#ifdef __cplusplus
}
#endif
